#include "loader.h"

#include "vector.h"

#include "mesh.h"
#include "twinedge.h"
#include "vertex.h"
#include "normal.h"
#include "material.h"
#include "group.h"
#include "vec3fv.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iostream.h>

FFloader::FFloader()
{
}

FFloader::~FFloader()
{
}

static void quicksort_for_connectivity(void** a, int i, int j)
{
    int in_i = i;
    int in_j = j;
    void* w;
    int small = (int)(((FFtwinEdge*)a[(i+j)/2])->boundary_prev);
    int large = (int)(((FFtwinEdge*)a[(i+j)/2])->boundary_next);
    do
    {
        while ( (((int)(((FFtwinEdge*)a[i])->boundary_next)) > large) || ((((int)(((FFtwinEdge*)a[i])->boundary_next)) == large) && (((int)(((FFtwinEdge*)a[i])->boundary_prev) > small))) ) i++;
        while ( (((int)(((FFtwinEdge*)a[j])->boundary_next)) < large) || ((((int)(((FFtwinEdge*)a[j])->boundary_next)) == large) && (((int)(((FFtwinEdge*)a[j])->boundary_prev) < small))) ) j--;
        if (i<j)
        {
            w = a[i];
            a[i] = a[j];
            a[j] = w;
        }
    } while (++i<=--j);
    if (i == j+3)
    {
        i--;
        j++;
    }
    if (j>in_i) quicksort_for_connectivity(a, in_i, j);
    if (i<in_j) quicksort_for_connectivity(a, i, in_j);
}

static void quicksort_for_materials(void** a, int i, int j)
{
    int in_i = i;
    int in_j = j;
    void* w;
    int key = ((FFtwinEdge*)a[(i+j)/2])->material;
    do
    {
        while ( ((FFtwinEdge*)a[i])->material < key ) i++;
	while ( ((FFtwinEdge*)a[j])->material > key ) j--;
        if (i<j)
        {
            w = a[i];
            a[i] = a[j];
            a[j] = w;
        }
    } while (++i<=--j);
    if (i == j+3)
    {
        i--;
        j++;
    }
    if (j>in_i) quicksort_for_materials(a, in_i, j);
    if (i<in_j) quicksort_for_materials(a, i, in_j);
}

int FFloader::loadObj(char* file_name, FFmesh* mesh)
{
  FILE *file;
  char line[255], *pointer;
  
  int i;
  float x, y, z;
  int v_idx, n_idx, t_idx;
  int has_normals = -1, has_textures = -1;
  FFtwinEdge *initEdge, *currEdge, *lastEdge;
  
  int corner_count = 0;
  
  if (file = fopen(file_name, "r"))
    {
      //      cout << "opened file '" << file_name << "'" << endl;
    }
  else
    {
      cout << "cannot open file '" << file_name << "'" << endl;
      return 0;
    }

  // we use this group index for all faces in case the model has no group structure    
  char current_group_name[128];
  strcpy(current_group_name, "no group");
  int current_group = mesh->gs->size();
  mesh->gs->addElement((void*) new FFgroup(current_group_name));
  
  // we use this material index for all faces in case the model has no material properties    
  char current_material_name[128];
  strcpy(current_material_name, "no material");
  int current_material = mesh->ms->size();
  mesh->ms->addElement((void*) new FFmaterial(current_group_name));
  
  while (fgets(line, sizeof(char) * 255, file))
    {
      if (*line == 'v')
	{
	  if (*(line+1) == 'n')
	    {
	      sscanf(line+3, "%f %f %f", &x, &y, &z);
	      mesh->ns->addElement((void*) new FFnormal(x, y, z));
	    }
	  else if (*(line+1) == ' ')
	    {
	      sscanf(line+2, "%f %f %f", &x, &y, &z);
	      mesh->vs->addElement((void*) new FFvertex(x, y, z));
	    }
	  else
	    {
	      /* textures and else are not implemented */
	    }
	}
      else if (*line == 'g')
	{
	  if (sscanf(line, "g %s", current_group_name) == 1) // prevent no-name groups
	    {
	      current_group = -1;
	      for (i = 0; i < mesh->gs->size(); i++)
		{
		  if (strcmp(&current_group_name[0], ((FFgroup*)mesh->gs->elementAt(i))->name) == 0)
		    {
		      current_group = i;
		    }
		}
	      if (current_group == -1)
		{
		  current_group = mesh->gs->size();
		  mesh->gs->addElement((void*) new FFgroup(current_group_name));
		}
	    }
	}
      else if (*line == 'f')
	{
	  lastEdge = 0;
	  
	  if (has_normals == -1)
	    {
	      has_normals = 0; has_textures = 0;
	      pointer = line + 1;
	      while ((*pointer  < '0') || (*pointer  > '9')) pointer++;
	      while ((*pointer >= '0') && (*pointer <= '9')) pointer++;
	      if (*pointer == '/')
		{
		  if (*(pointer+1) == '/')
		    {
		      has_normals = 1;
		    }
		  else
		    {
		      has_normals = 1;
		      has_textures = 1;
		    }
		}
	    }

	  pointer = line+1;
	  while ((*pointer  < '0') || (*pointer  > '9')) pointer++;
	  while ((*pointer >= '0') && (*pointer <= '9')) pointer++;

	  // now read in the face
	  pointer = line+1;
	  
	  while (1)
	    {
	      while ((*pointer != 0) && ((*pointer < '0') || (*pointer > '9'))) pointer++;
	      
	      if (*pointer == 0)
		{
		  break;
		}

	      corner_count++;
	      
	      if ((has_normals == 0) && (has_textures == 0))
		{
		  sscanf(pointer, "%d", &v_idx);
		  currEdge = new FFtwinEdge(current_group, current_material, (FFvertex*)(v_idx-1));
		}
	      else if ((has_normals == 1) && (has_textures == 0))
		{
		  sscanf(pointer, "%d//%d", &v_idx, &n_idx);
		  currEdge = new FFtwinEdge(current_group, current_material, (FFvertex*)(v_idx-1), (FFnormal*)(n_idx-1));
		}
	      else if ((has_normals == 0) && (has_textures == 1))
		{
		  sscanf(pointer, "%d/%d", &v_idx, &t_idx);
		  currEdge = new FFtwinEdge(current_group, current_material, (FFvertex*)(v_idx-1));
		}
	      else if ((has_normals == 1) && (has_textures == 1))
		{
		  sscanf(pointer, "%d/%d/%d", &v_idx, &n_idx, &t_idx);
		  currEdge = new FFtwinEdge(current_group, current_material, (FFvertex*)(v_idx-1), (FFnormal*)(n_idx-1));
		}
	      
	      while ((*pointer == '/') || ((*pointer >= '0') && (*pointer <= '9'))) pointer++;
	      
	      if (lastEdge)
		{
		  currEdge->next = lastEdge;
		  if (((int)lastEdge->vertex) > ((int)currEdge->vertex))
		    {
		      lastEdge->boundary_prev = (FFtwinEdge*)currEdge->vertex; /* used as key in quicksort */
		      lastEdge->boundary_next = (FFtwinEdge*)lastEdge->vertex; /* used as key in quicksort */
		    }
		  else
		    {
		      lastEdge->boundary_prev = (FFtwinEdge*)lastEdge->vertex; /* used as key in quicksort */
		      lastEdge->boundary_next = (FFtwinEdge*)currEdge->vertex; /* used as key in quicksort */
		    }
		}
	      else
		{
		  initEdge = currEdge;
		}
	      mesh->es->addElement(currEdge);
	      lastEdge = currEdge;
	    }
	  mesh->fs->addElement(initEdge);
	  initEdge->next = lastEdge;
	  if (((int)lastEdge->vertex) > ((int)initEdge->vertex))
	    {
	      lastEdge->boundary_prev = (FFtwinEdge*)initEdge->vertex; /* used as key in quicksort */
	      lastEdge->boundary_next = (FFtwinEdge*)lastEdge->vertex; /* used as key in quicksort */
	    }
	  else
	    {
	      lastEdge->boundary_prev = (FFtwinEdge*)lastEdge->vertex; /* used as key in quicksort */
	      lastEdge->boundary_next = (FFtwinEdge*)initEdge->vertex; /* used as key in quicksort */
	    }
	}
      else if ((*line == 'u') && (*(line+1) == 's') && (*(line+2) == 'e'))
	{
	  sscanf(line+6, "%s", current_material_name);	  
	  current_material = -1;
	  for (i = 0; i < mesh->ms->size(); i++)
	    {
	      if (strcmp(current_material_name, ((FFmaterial*)mesh->ms->elementAt(i))->name) == 0)
		{
		  current_material = i;
		  break;
		}
	    }
	  if (current_material == -1)
	    {
	      current_material = mesh->ms->size()-1;
	    }
	} 
    }
  
  cout << "# of vertices  : " << mesh->vs->size() << endl;
  cout << "# of normals   : " << mesh->ns->size() << endl;
  cout << "# of groups    : " << mesh->gs->size()-1 << endl;
  cout << "# of faces     : " << mesh->fs->size() << endl;
  cout << "# of corners   : " << corner_count << endl;
  cout << "# of i-edges   : " << mesh->es->size() << endl;
  
  quicksort_for_connectivity( mesh->es->data, 0, mesh->es->size()-1 );

  quicksort_for_materials( mesh->fs->data, 0, mesh->fs->size()-1 );
  
  int non_boundary_edges = mesh->es->size();
  
  for (i = 0; i < non_boundary_edges; i++)
    {
      initEdge = (FFtwinEdge*)mesh->es->data[i];
      currEdge = (FFtwinEdge*)mesh->es->data[i+1];
      
      if ((initEdge->boundary_prev == currEdge->boundary_prev) && (initEdge->boundary_next == currEdge->boundary_next))
	{
	  initEdge->inv = currEdge;
	  currEdge->inv = initEdge;
	  i++;
	}
      else
	{
	  lastEdge = new FFtwinEdge(-1, -1); // group index -1 means hole, vertex idx not known here
	  mesh->es->addElement(lastEdge);
	  initEdge->inv = lastEdge;
	  lastEdge->inv = initEdge;
	}
    }
  
  int boundary_edges = mesh->es->size();
  
  for (i = 0; i < non_boundary_edges; i++)
    {
      currEdge = (FFtwinEdge*)mesh->es->data[i];
      currEdge->next = currEdge->next->inv;
      currEdge->vertex = (FFvertex*)mesh->vs->elementAt((int)currEdge->vertex);
      if (has_normals) currEdge->normal = (FFnormal*)mesh->ns->elementAt((int)currEdge->normal);
      else             currEdge->normal = (FFnormal*)0;
      currEdge->boundary_prev = 0;
      currEdge->boundary_next = 0;
    }
  
  cout << "# of b-edges   : " << (boundary_edges-non_boundary_edges) << endl;
  
  for (i = non_boundary_edges; i < boundary_edges; i++)
    {
      currEdge = (FFtwinEdge*)mesh->es->elementAt(i);      
      initEdge = currEdge->inv->next;
      while (initEdge->inv->next != currEdge)
	{
	  initEdge = initEdge->inv->next;
	}
      lastEdge = initEdge->inv;
      while (lastEdge->inv->group != -1)
	{
	  initEdge = lastEdge->inv->next;
	  while (initEdge->inv->next != lastEdge)
	    {
	      initEdge = initEdge->inv->next;
	    }
	  lastEdge = initEdge->inv;
	}
      currEdge->next = lastEdge;
      currEdge->vertex = lastEdge->vertex;
      currEdge->normal = lastEdge->normal;
    }
  return 1;
}

int FFloader::loadMtl(char* file_name, FFmesh* mesh)
{
  FILE *file;
  char line[255];  
  float r, g, b;
  FFmaterial* m;
  
  if (file = fopen(file_name, "r"))
    {
      //      cout << "opened file '" << file_name << "'" << endl;
    }
  else
    {
      cout << "cannot open file '" << file_name << "'" << endl;
    }
  
  while (fgets(line, sizeof(char) * 255, file))
    {
      if (*line == '\n')
	{
	  // empty line
	}
      else if ((*line == 'n') && (*(line+1) == 'e') && (*(line+2) == 'w'))
	{
	  m = new FFmaterial();
	  sscanf(line+6, "%s", m->name);
	  mesh->ms->addElement((void*) m);
	}
      else if (*line == 'K')
	{
	  sscanf(line+3, "%f %f %f", &r, &g, &b);

	  if ((*(line+1) == 'a'))
	    {
	      VecSet3fv(m->a, r, g, b);
	    }
	  else if ((*(line+1) == 'd'))
	    {
	      VecSet3fv(m->d, r, g, b);
	    }
	  else if ((*(line+1) == 's'))
	    {
	      VecSet3fv(m->s, r, g, b);
	    }
	  else if ((*(line+1) == 'e'))
	    {
	      VecSet3fv(m->e, r, g, b);
	    }
	  else
	    {
	      cout << "cannot understand: " << line << endl;
	    }
	}
      else 
	{
	  cout << "cannot understand: " << line << endl;
	}
    }
  cout << "# of materials : " << mesh->ms->size() << endl;
  return 1;
}
