/*
===============================================================================

	FILE:	main.cpp
	
	CONTENTS:
	
		Face Fixer - SIGGRAPH 2000 demonstration software
	
	PROGRAMMERS:
	
		Martin Isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		28 April 2000 -- Martin finalized for SIGGRAPH submission.
	
===============================================================================
*/
#include <iostream.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <GL/glut.h>

#include "loader.h"
#include "encoder.h"
#include "mesh.h"
#include "codec.h"
#include "label.h"
#include "vertex.h"
#include "normal.h"
#include "twinedge.h"
#include "material.h"
#include "vector.h"
#include "slider.h"

// MOUSE INTERFACE
int LeftButtonDown=0;        
int MiddleButtonDown=0;        
int RightButtonDown=0;
int OldX,OldY,NewX,NewY;
float Elevation=0;
float Azimuth=0;
float DistX=0;
float DistY=0;
float DistZ=3;

// GLOBAL CONTROL VARIABLES
int WindowW=640, WindowH=480;
int LightingAnimationOn=0;
int LightCount=5;
int CodingAnimationOn=0;
int CodingAnimationSpeed=0;
int WireFrameOn=0;
int VertexNormalsOn=0;
int InteractionMode=0;
int ColourMode=0;
int ViewMode=0;
int ViewModeGroup=1;
int ViewModeComponent=1;

// FOR SLIDER
TUISlider* aSlider;
int WindowHeight()
{
  return WindowH;
}

// GLOBAL FACE FIXER VARIABLES
FFloader *loader = 0;
FFmesh *mesh = 0;
FFencoder *encoder = 0;
FFcodec *full_codec = 0;
FFcodec *step_codec = 0;

// GLOBAL MATERIAL VARIABLES
float colours_ambient[10][4];
float colours_diffuse[10][4];
float colours_specular[4][4];
float colours_emission[5][4];

// ANIMATED LIGHTS
float Light0pos[]={0,-1,0,1}, d0[]={-0.05f,0.15f,0.125f};
float Light1pos[]={1,1,0,1},  d1[]={0.13f,-0.06f,0.16f};
float Light2pos[]={-1,1,0,1}, d2[]={-0.065f,0.16f,-0.1f};
float Light3pos[]={-1,-1,0,1}, d3[]={0.05f,-0.15f,0.125f};
float Light4pos[]={1,1,1,1},  d4[]={-0.13f,-0.06f,-0.16f};
float Light5pos[]={-1,-1,-1,1}, d5[]={-0.15f,-0.13f,-0.1f};
float Light6pos[]={-1,-1,1,1}, d6[]={0.1f,0.08f,-0.05f};
float LightBounds[]={1.75,1.75,1.75};

void AnimateLights()
{
  for (int i=0; i<3; i++)
  {
    Light0pos[i]+=d0[i];
    Light1pos[i]+=d1[i];
    Light2pos[i]+=d2[i];
    Light3pos[i]+=d3[i];
    Light4pos[i]+=d4[i];
    Light5pos[i]+=d5[i];
    Light6pos[i]+=d6[i];
    if (Light0pos[i]<-LightBounds[i] || Light0pos[i]>LightBounds[i]) 
      { d0[i]=-d0[i]; Light0pos[i]+=d0[i]; }
    if (Light1pos[i]<-LightBounds[i] || Light1pos[i]>LightBounds[i]) 
      { d1[i]=-d1[i]; Light1pos[i]+=d1[i]; }
    if (Light2pos[i]<-LightBounds[i] || Light2pos[i]>LightBounds[i]) 
      { d2[i]=-d2[i]; Light2pos[i]+=d2[i]; }
    if (Light3pos[i]<-LightBounds[i] || Light3pos[i]>LightBounds[i]) 
      { d3[i]=-d3[i]; Light3pos[i]+=d3[i]; }
    if (Light4pos[i]<-LightBounds[i] || Light4pos[i]>LightBounds[i]) 
      { d4[i]=-d4[i]; Light4pos[i]+=d4[i]; }
    if (Light5pos[i]<-LightBounds[i] || Light5pos[i]>LightBounds[i]) 
      { d5[i]=-d5[i]; Light5pos[i]+=d5[i]; }
    if (Light6pos[i]<-LightBounds[i] || Light6pos[i]>LightBounds[i]) 
      { d6[i]=-d6[i]; Light6pos[i]+=d6[i]; }
  }
}

void encodeMeshRel(int rel)
{
  if (rel > 0)
    {
      if (step_codec == 0)
	{
	  FFtwinEdge* e;
	  step_codec = new FFcodec();
	  for (int i = 0; i < mesh->es->size(); i++)
	    {
	      e = (FFtwinEdge*) mesh->es->elementAt(i);
	      e->boundary_prev = 0;
	      e->boundary_next = 0;
	    }
	  mesh->gate = (FFtwinEdge*) mesh->es->elementAt(0);
	}
      encoder->encodeStepwise(mesh, step_codec, rel);
    }
  if (step_codec)
    {
      aSlider->SetValue(step_codec->ls->size());
    }
}

void encodeMeshAbs(int abs)
{
  if (step_codec == 0)
    {
      encodeMeshRel(abs);
    }
  else
    {
      encodeMeshRel(abs - step_codec->ls->size());
    }
}

void displayMessage( )
{
  glColor3f( 0.7, 0.7, 0.7 );  // Set colour to grey
  glMatrixMode( GL_PROJECTION );
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D( 0.0, 1.0, 0.0, 1.0 );
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();
  glRasterPos2f( 0.03, 0.95 );

  if( InteractionMode == 0 )
    {
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
    }
  else if( InteractionMode == 1 )
    {    
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'r');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'n');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 's');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'l');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'a');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 't');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'e');
    }
  else
    {
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'z');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'o');
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, 'm');
    }

    glPopMatrix();
    glMatrixMode( GL_PROJECTION );
    glPopMatrix();
}

void myDisplay()
{  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
  glViewport(0,0,WindowW,WindowH);
  glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(30,(float)WindowW/WindowH,0.125f,10);
  glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(DistX,DistY,DistZ, DistX,DistY,0, 0,1,0);
    glRotatef(Elevation,1,0,0);
    glRotatef(Azimuth,0,1,0);

    glLightfv(GL_LIGHT0,GL_POSITION,Light0pos);
    if (LightCount > 1) glLightfv(GL_LIGHT1,GL_POSITION,Light1pos);
    if (LightCount > 2) glLightfv(GL_LIGHT2,GL_POSITION,Light2pos);
    if (LightCount > 3) glLightfv(GL_LIGHT3,GL_POSITION,Light3pos);
    if (LightCount > 4) glLightfv(GL_LIGHT4,GL_POSITION,Light4pos);
    if (LightCount > 5) glLightfv(GL_LIGHT5,GL_POSITION,Light5pos);
    if (LightCount > 6) glLightfv(GL_LIGHT6,GL_POSITION,Light6pos);

  if (ViewMode == 0)
    {
      glEnable(GL_LIGHTING);
      glEnable(GL_CULL_FACE);  

      if (ColourMode == 0) // use material colours
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_material = -1;
	  FFmaterial *m;
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->material != current_material)
		    {
		      current_material = e0->material;
		      m = (FFmaterial*) mesh->ms->elementAt(current_material);
		      glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		      glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		      glMaterialfv(GL_FRONT,GL_EMISSION,m->e);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 1) // colours by group
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_group = -1;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[2]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->group != current_group)
		    {
		      current_group = e0->group;
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_group%8)+1]);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 2) // colours by component
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_component = -1;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[2]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->component != current_component)
		    {
		      current_component = e0->component;
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_component%8)+1]);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 3) // mark triangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (triangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		    }
		}
	      }
	  glEnd();
	  // draw faces that are not encoded yet (non-triangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 4) // mark quadrangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (quadrangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-quadrangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 5) // mark pentagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (pentagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-pentagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 6) // mark hexagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (hexagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if ((e0->next->inv->next->inv->next->inv != e0) && (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv == e0))
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-hexagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if ((e0->next->inv->next->inv->next->inv == e0) || (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv != e0))
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 7) // mark biggergons
	{
	  int i, count;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->marker)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (biggergons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  e1 = e0->next->inv->next->inv->next->inv;
		  count = 0;
		  while (e1 != e0)
		    {
		      e1 = e1->next->inv;
		      count++;
		    }
		  if (count > 3)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  // draw faces that are not encoded yet (non-biggergons)
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  e1 = e0->next->inv->next->inv->next->inv;
		  count = 0;
		  while (e1 != e0)
		    {
		      e1 = e1->next->inv;
		      count++;
		    }
		  if (count <= 3)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      
      glDisable(GL_LIGHTING);
      glDisable(GL_CULL_FACE);  
      
      if ( VertexNormalsOn == 1 ) // all vertex normals
	{
	  FFtwinEdge *e0;
	  glColor3f(1.0, 1.0, 1.0);
	  glLineWidth(1.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      glVertex3fv(e0->vertex->v);
	      glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.05,e0->vertex->v[1]+e0->normal->n[1]*0.05,e0->vertex->v[2]+e0->normal->n[2]*0.05);
	    }
	  glEnd();
	}
      else if ( VertexNormalsOn == 2 ) //  only single vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      normal = e0->normal;
	      e1 = e0->next;
	      while (e1 != e0)
		{
		  if (normal != e1->normal)
		    {
		      normal = 0;
		      break;
		    }
		  e1 = e1->next;
		}
	      if (normal)
		{
		  glVertex3fv(e0->vertex->v);
		  glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
		}
	    }
	  glEnd();      
	}
      else if ( VertexNormalsOn == 3 ) //  only multiple vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      normal = e0->normal;
	      e1 = e0->next;	  
	      while (e1 != e0)
		{
		  if (normal != e1->normal)
		    {
		      normal = e1->normal;
		      glVertex3fv(e1->vertex->v);
		      glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
		    }
		  e1 = e1->next;
		}
	    }
	  glEnd();      
	}
      
      if ( WireFrameOn )
	{
	  FFtwinEdge *e0;
	  glColor3f(0.0, 0.0, 0.0);
	  glLineWidth(1.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
	      glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.001,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.001,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.001);
	    }
	  glEnd();
	}

      if (step_codec)
	{
	  FFtwinEdge *e0, *e1;
	  if (mesh->gate)
	    {
	      e0 = mesh->gate->boundary_next;
	      glColor3f(0.5, 0.0, 0.0);
	      glLineWidth(6.0);
	      glBegin(GL_LINES);
	      while (e0 != mesh->gate)
		{
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.0011,e0->vertex->v[1]+e0->normal->n[1]*0.0011,e0->vertex->v[2]+e0->normal->n[2]*0.0011);
		  glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.0011,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.0011,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.0011);
		  e0 = e0->boundary_next;
		}	  
	      glEnd();
	      glColor3f(1.0, 0.2, 0.2);
	      glLineWidth(10.0);
	      glBegin(GL_LINES);
	      glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.0012,e0->vertex->v[1]+e0->normal->n[1]*0.0012,e0->vertex->v[2]+e0->normal->n[2]*0.0012);
	      glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.0012,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.0012,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.0012);
	      glEnd();	      
	    }
	  for (int i = 0; i < step_codec->gs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) step_codec->gs->elementAt(i);
	      e1 = e0->boundary_next;
	      glColor3f(0.2, 0.2, 0.5);
	      glLineWidth(3.0);
	      glBegin(GL_LINES);
	      while (e1 != e0)
		{
		  glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.00105,e1->vertex->v[1]+e1->normal->n[1]*0.00105,e1->vertex->v[2]+e1->normal->n[2]*0.00105);
		  glVertex3f(e1->inv->vertex->v[0]+e1->inv->normal->n[0]*0.00105,e1->inv->vertex->v[1]+e1->inv->normal->n[1]*0.00105,e1->inv->vertex->v[2]+e1->inv->normal->n[2]*0.00105);
		  e1 = e1->boundary_next;
		}
	      glEnd();
	      glColor3f(0.5, 0.5, 1.0);
	      glLineWidth(9.0);
	      glBegin(GL_LINES);
	      glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.00115,e0->vertex->v[1]+e0->normal->n[1]*0.00115,e0->vertex->v[2]+e0->normal->n[2]*0.00115);
	      glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.00115,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.00115,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.00115);
	      glEnd();	      
	    }
	}
    }
  else if (ViewMode == 1)  // show only one group
    {
      glEnable(GL_LIGHTING);
      
      if (ColourMode == 0) // use material colours
	{
	  int current_material = -1;
	  FFmaterial *m;
	  FFtwinEdge *e0, *e1, *e2;
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->material != current_material)
		    {
		      current_material = e0->material;
		      m = (FFmaterial*) mesh->ms->elementAt(current_material);
		      glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
		      glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
		      glMaterialfv(GL_FRONT,GL_EMISSION,m->e);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 1) // colours by group
	{
	  int current_group = -1;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->group != current_group)
		    {
		      current_group = e0->group;
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_group%8)+1]);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 2) // colours by component
	{
	  int current_component = -1;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->component != current_component)
		    {
		      current_component = e0->component;
		      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_component%8)+1]);
		    }
		  e1 = e0->next->inv;
		  e2 = e1->next->inv;
		  do
		    {
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
		    }
		  while (e1 != e0);
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 3) // mark triangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    } 
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 4) // mark quadrangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 5) // mark pentagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv->next->inv == e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if (e0->next->inv->next->inv->next->inv->next->inv->next->inv != e0)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 6) // mark hexagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if ((e0->next->inv->next->inv->next->inv != e0) && (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv == e0))
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  if ((e0->next->inv->next->inv->next->inv == e0) || (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv != e0))
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 7) // mark biggergons
	{
	  int i, count;
	  FFtwinEdge *e0, *e1, *e2;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  e1 = e0->next->inv->next->inv->next->inv;
		  count = 0;
		  while (e1 != e0)
		    {
		      e1 = e1->next->inv;
		      count++;
		    }
		  if (count > 3)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  e1 = e0->next->inv->next->inv->next->inv;
		  count = 0;
		  while (e1 != e0)
		    {
		      e1 = e1->next->inv;
		      count++;
		    }
		  if (count <= 3)
		    {
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0); 
		    }
		}
	    }
	  glEnd();
	}
      
      glDisable(GL_LIGHTING);
      
      if ( VertexNormalsOn == 1 ) // all vertex normals
	{
	  FFtwinEdge *e0;
	  glColor3f(1.0, 1.0, 1.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  glVertex3fv(e0->vertex->v);
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.05,e0->vertex->v[1]+e0->normal->n[1]*0.05,e0->vertex->v[2]+e0->normal->n[2]*0.05);
		}
	    }
	  glEnd();
	}
      else if ( VertexNormalsOn == 2 ) //  only single vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  normal = e0->normal;
		  e1 = e0->next;
		  while (e1 != e0)
		    {
		      if (normal != e1->normal)
			{
			  normal = 0;
			  break;
			}
		      e1 = e1->next;
		    }
		  if (normal)
		    {
		      glVertex3fv(e0->vertex->v);
		      glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
		    }
		}
	    }
	  glEnd();      
	}
      else if ( VertexNormalsOn == 3 ) //  only multiple vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  normal = e0->normal;
		  e1 = e0->next;	  
		  while (e1 != e0)
		    {
		      if (normal != e1->normal)
			{
			  normal = e1->normal;
			  glVertex3fv(e1->vertex->v);
			  glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
			}
		      e1 = e1->next;
		    }
		}
	    }
	  glEnd();      
	}

      if ( WireFrameOn )
	{
	  FFtwinEdge *e0;
	  glColor3f(0.0, 0.0, 0.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->group == ViewModeGroup)
		{
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
		  glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.001,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.001,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.001);
		}
	    }
	  glEnd();
	}
    }
  else if (ViewMode == 2)
    {
      glEnable(GL_LIGHTING);
      
      if (ColourMode == 0) // use material colours
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_material = -1;
	  FFmaterial *m;
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->marker == 0)
		{
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->material != current_material)
			{
			  current_material = e0->material;
			  m = (FFmaterial*) mesh->ms->elementAt(current_material);
			  glMaterialfv(GL_FRONT,GL_AMBIENT,m->a);
			  glMaterialfv(GL_FRONT,GL_DIFFUSE,m->d);
			  glMaterialfv(GL_FRONT,GL_SPECULAR,m->s);
			  glMaterialfv(GL_FRONT,GL_EMISSION,m->e);
			}
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 1) // colours by group
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_group = -1;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[2]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->group != current_group)
			{
			  current_group = e0->group;
			  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_group%8)+1]);
			}
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
		      glNormal3fv(e0->normal->n);
		      glVertex3fv(e0->vertex->v);
		      glNormal3fv(e2->normal->n);
		      glVertex3fv(e2->vertex->v);
		      glNormal3fv(e1->normal->n);
		      glVertex3fv(e1->vertex->v);
		      e1 = e1->next->inv;
		      e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 2) // colours by component
	{
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (int i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet
	  int current_component = -1;
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[0]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[2]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (int i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->component != current_component)
			{
			  current_component = e0->component;
			  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[(current_component%8)+1]);
			}
		      e1 = e0->next->inv;
		      e2 = e1->next->inv;
		      do
			{
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			  e1 = e1->next->inv;
			  e2 = e2->next->inv;
			}
		      while (e1 != e0);
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 3) // mark triangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (triangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv == e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  glNormal3fv(e0->normal->n);
			  glVertex3fv(e0->vertex->v);
			  glNormal3fv(e2->normal->n);
			  glVertex3fv(e2->vertex->v);
			  glNormal3fv(e1->normal->n);
			  glVertex3fv(e1->vertex->v);
			}
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-triangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv != e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			} 
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 4) // mark quadrangles
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (quadrangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv->next->inv == e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-quadrangles)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv->next->inv != e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 5) // mark pentagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (pentagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv->next->inv->next->inv == e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-pentagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if (e0->next->inv->next->inv->next->inv->next->inv->next->inv != e0)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 6) // mark hexagons
	{
	  int i;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (hexagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if ((e0->next->inv->next->inv->next->inv != e0) && (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv == e0))
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-hexagons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      if ((e0->next->inv->next->inv->next->inv == e0) || (e0->next->inv->next->inv->next->inv->next->inv->next->inv->next->inv != e0))
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			}
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	}
      else if (ColourMode == 7) // mark biggergons
	{
	  int i, count;
	  FFtwinEdge *e0, *e1, *e2;
	  // draw faces that are already encoded
	  if (step_codec)
	    {
	      glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[9]);
	      glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[9]);
	      glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[3]);
	      glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[4]);
	      glBegin(GL_TRIANGLES);
	      for (i = 0; i < mesh->fs->size(); i++)
		{
		  e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
		  if (e0->component == ViewModeComponent)
		    {
		      if (e0->marker)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0);
			}
		    }
		}
	      glEnd();
	    }
	  // draw faces that are not encoded yet (biggergons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[1]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[1]);
	  glMaterialfv(GL_FRONT,GL_SPECULAR,colours_specular[1]);
	  glMaterialfv(GL_FRONT,GL_EMISSION,colours_emission[0]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      e1 = e0->next->inv->next->inv->next->inv;
		      count = 0;
		      while (e1 != e0)
			{
			  e1 = e1->next->inv;
			  count++;
			}
		      if (count > 3)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	  // draw faces that are not encoded yet (non-biggergons)
	  glMaterialfv(GL_FRONT,GL_AMBIENT,colours_ambient[2]);
	  glMaterialfv(GL_FRONT,GL_DIFFUSE,colours_diffuse[2]);
	  glBegin(GL_TRIANGLES);
	  for (i = 0; i < mesh->fs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->fs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  if (e0->marker == 0)
		    {
		      e1 = e0->next->inv->next->inv->next->inv;
		      count = 0;
		      while (e1 != e0)
			{
			  e1 = e1->next->inv;
			  count++;
			}
		      if (count <= 3)
			{
			  e1 = e0->next->inv;
			  e2 = e1->next->inv;
			  do
			    {
			      glNormal3fv(e0->normal->n);
			      glVertex3fv(e0->vertex->v);
			      glNormal3fv(e2->normal->n);
			      glVertex3fv(e2->vertex->v);
			      glNormal3fv(e1->normal->n);
			      glVertex3fv(e1->vertex->v);
			      e1 = e1->next->inv;
			      e2 = e2->next->inv;
			    }
			  while (e1 != e0); 
			}
		    }
		}
	    }
	  glEnd();
	}
      
      glDisable(GL_LIGHTING);
      
      if ( VertexNormalsOn == 1 ) // all vertex normals
	{
	  FFtwinEdge *e0;
	  glColor3f(1.0, 1.0, 1.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  glVertex3fv(e0->vertex->v);
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.05,e0->vertex->v[1]+e0->normal->n[1]*0.05,e0->vertex->v[2]+e0->normal->n[2]*0.05);
		}
	    }
	  glEnd();
	}
      else if ( VertexNormalsOn == 2 ) //  only single vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  normal = e0->normal;
		  e1 = e0->next;
		  while (e1 != e0)
		    {
		      if (normal != e1->normal)
			{
			  normal = 0;
			  break;
			}
		      e1 = e1->next;
		    }
		  if (normal)
		    {
		      glVertex3fv(e0->vertex->v);
		      glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
		    }
		}
	    }
	  glEnd();      
	}
      else if ( VertexNormalsOn == 3 ) //  only multiple vertex normals
	{
	  FFtwinEdge *e0, *e1;
	  glColor3f(1.0,1.0,1.0);
	  glLineWidth(1.0);
	  FFnormal* normal;
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  normal = e0->normal;
		  e1 = e0->next;	  
		  while (e1 != e0)
		    {
		      if (normal != e1->normal)
			{
			  normal = e1->normal;
			  glVertex3fv(e1->vertex->v);
			  glVertex3f(e0->vertex->v[0]+normal->n[0]*0.05,e0->vertex->v[1]+normal->n[1]*0.05,e0->vertex->v[2]+normal->n[2]*0.05);
			}
		      e1 = e1->next;
		    }
		}
	    }
	  glEnd();      
	}
      
      if ( WireFrameOn )
	{
	  FFtwinEdge *e0;
	  glColor3f(0.0, 0.0, 0.0);
	  glBegin(GL_LINES);
	  for  (int i = 0; i < mesh->es->size(); i++)
	    {
	      e0 = (FFtwinEdge*) mesh->es->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.001,e0->vertex->v[1]+e0->normal->n[1]*0.001,e0->vertex->v[2]+e0->normal->n[2]*0.001);
		  glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.001,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.001,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.001);
		}
	    }
	  glEnd();
	}
      if (step_codec)
	{
	  FFtwinEdge *e0, *e1;
	  if (mesh->gate)
	    {
	      if (mesh->gate->component == ViewModeComponent)
		{
		  e0 = mesh->gate->boundary_next;
		  glColor3f(0.5, 0.0, 0.0);
		  glLineWidth(6.0);
		  glBegin(GL_LINES);
		  while (e0 != mesh->gate)
		    {
		      glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.0011,e0->vertex->v[1]+e0->normal->n[1]*0.0011,e0->vertex->v[2]+e0->normal->n[2]*0.0011);
		      glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.0011,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.0011,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.0011);
		      e0 = e0->boundary_next;
		    }	  
		  glEnd();
		  glColor3f(1.0, 0.2, 0.2);
		  glLineWidth(10.0);
		  glBegin(GL_LINES);
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.0012,e0->vertex->v[1]+e0->normal->n[1]*0.0012,e0->vertex->v[2]+e0->normal->n[2]*0.0012);
		  glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.0012,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.0012,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.0012);
		  glEnd();	      
		}
	    }
	  for (int i = 0; i < step_codec->gs->size(); i++)
	    {
	      e0 = (FFtwinEdge*) step_codec->gs->elementAt(i);
	      if (e0->component == ViewModeComponent)
		{
		  
		  e1 = e0->boundary_next;
		  glColor3f(0.2, 0.2, 0.5);
		  glLineWidth(3.0);
		  glBegin(GL_LINES);
		  while (e1 != e0)
		    {
		      glVertex3f(e1->vertex->v[0]+e1->normal->n[0]*0.00105,e1->vertex->v[1]+e1->normal->n[1]*0.00105,e1->vertex->v[2]+e1->normal->n[2]*0.00105);
		      glVertex3f(e1->inv->vertex->v[0]+e1->inv->normal->n[0]*0.00105,e1->inv->vertex->v[1]+e1->inv->normal->n[1]*0.00105,e1->inv->vertex->v[2]+e1->inv->normal->n[2]*0.00105);
		      e1 = e1->boundary_next;
		    }
		  glEnd();
		  glColor3f(0.5, 0.5, 1.0);
		  glLineWidth(9.0);
		  glBegin(GL_LINES);
		  glVertex3f(e0->vertex->v[0]+e0->normal->n[0]*0.00115,e0->vertex->v[1]+e0->normal->n[1]*0.00115,e0->vertex->v[2]+e0->normal->n[2]*0.00115);
		  glVertex3f(e0->inv->vertex->v[0]+e0->inv->normal->n[0]*0.00115,e0->inv->vertex->v[1]+e0->inv->normal->n[1]*0.00115,e0->inv->vertex->v[2]+e0->inv->normal->n[2]*0.00115);
		  glEnd();	      
		}
	    }
	}
    }
  
  displayMessage( );
  
  glDisable(GL_DEPTH_TEST);
  aSlider->DoDisplay();
  glEnable(GL_DEPTH_TEST);  
  
  glutSwapBuffers();
}

void myReshape(int w, int h)
{
  glutReshapeWindow(WindowW,WindowH);
}

void myIdle()
{
  if (LightingAnimationOn)
    {
      AnimateLights();
    }
  if (CodingAnimationOn)
    {
      if (CodingAnimationSpeed == 0)
	{
	  encodeMeshRel(1);
	}
      else if (CodingAnimationSpeed == 1)
	{
	  encodeMeshRel(2);
	}
      else if (CodingAnimationSpeed == 2)
	{
	  encodeMeshRel(3);
	}
    }
  glutPostRedisplay();
}

void myMouseFunc(int button, int state, int x, int y)
{
  NewX=x; NewY=y;

  if ( aSlider->IsInside(x,y) || aSlider->IsDragging() )
    {
      if (state == GLUT_DOWN)
	{
	  aSlider->OnMouseDown(button, x, y);
	}
      else if (state == GLUT_UP)
	{
	  aSlider->OnMouseUp(button, x, y);
	  encodeMeshAbs(aSlider->GetValue());
	}
    }

  if (button == GLUT_LEFT_BUTTON)
    {
      LeftButtonDown =! LeftButtonDown;
    }
  else if (button == GLUT_RIGHT_BUTTON)
    {
      RightButtonDown =! RightButtonDown;
    }
  else if (button == GLUT_MIDDLE_BUTTON)
    {
      MiddleButtonDown =! MiddleButtonDown;
    }
}

void myMotionFunc(int x, int y)
{
  OldX=NewX; OldY=NewY;
  NewX=x;    NewY=y;

  if (aSlider->IsInside(x,y) || aSlider->IsDragging())
    {
      aSlider->OnMouseMove(x, y);
      encodeMeshAbs(aSlider->GetValue());
      return;
    }
  
  float RelX = (NewX-OldX) / (float)glutGet((GLenum)GLUT_WINDOW_WIDTH);
  float RelY = (NewY-OldY) / (float)glutGet((GLenum)GLUT_WINDOW_HEIGHT);
  if (LeftButtonDown) 
    { 
      if (InteractionMode == 0)
	{
	  Azimuth += (RelX*180);
	  Elevation += (RelY*180);
	}
      else if (InteractionMode == 1)
	{
	  DistX-=RelX;
	  DistY+=RelY;
	}
      else if (InteractionMode == 2)
	{
	  DistZ-=RelY*DistZ;
	}
    }
  else if (MiddleButtonDown)
    {
      DistX-=RelX*1.0f;
      DistY+=RelY*1.0f;
    }
}

void InitLights()
{
  float Lightdiff[] = {1,1,1,1};
  glLightfv(GL_LIGHT0,GL_DIFFUSE,Lightdiff);
  glEnable(GL_LIGHT0);
  if (LightCount > 1)
    {
      glLightfv(GL_LIGHT1,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT1);
    }
  else
    {
      glDisable(GL_LIGHT1);
    }
  if (LightCount > 2)
    {
      glLightfv(GL_LIGHT2,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT2);
    }
  else
    {
      glDisable(GL_LIGHT2);
    }
  if (LightCount > 3)
    {
      glLightfv(GL_LIGHT3,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT3);
    }
  else
    {
      glDisable(GL_LIGHT3);
    }
  if (LightCount > 4)
    {
      glLightfv(GL_LIGHT4,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT4);
    }
  else
    {
      glDisable(GL_LIGHT4);
    }
  if (LightCount > 5)
    {
      glLightfv(GL_LIGHT5,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT5);
    }
  else
    {
      glDisable(GL_LIGHT5);
    }
  if (LightCount > 6)
    {
      glLightfv(GL_LIGHT6,GL_DIFFUSE,Lightdiff);
      glEnable(GL_LIGHT6);
    }
  else
    {
      glDisable(GL_LIGHT6);
    }
}

void MyMenuFunc(int value)
{
  if (value == 1)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("teapot_solid.obj");
      loader->loadMtl("teapot.mtl", mesh);
      loader->loadObj("teapot.obj", mesh);
      mesh->normalize();
      mesh->computePerVertexNormals();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 2)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("cow_meat_cuts.obj");
      loader->loadMtl("cow_w_mc.mtl", mesh);
      loader->loadObj("cow_w_mc.obj", mesh);
      mesh->normalize();
      mesh->computePerVertexNormals();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 3)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("triceratops.obj");
      loader->loadMtl("trictops.mtl", mesh);
      loader->loadObj("trictops.obj", mesh);
      mesh->normalize();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
   }
  else if (value == 4)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("galleon.obj");
      loader->loadMtl("galleon.mtl", mesh);
      loader->loadObj("galleon.obj", mesh);
      mesh->normalize();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 5)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("shark.obj");
      loader->loadMtl("shark.mtl", mesh);
      loader->loadObj("shark.obj", mesh);
      mesh->normalize();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 6)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("beethoven.obj");
      loader->loadMtl("beethven.mtl", mesh);
      loader->loadObj("beethven.obj", mesh);
      mesh->normalize();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 7)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("cupie.obj");
      loader->loadMtl("cupie.mtl", mesh);
      loader->loadObj("cupie.obj", mesh);
      mesh->normalize();
      mesh->computePerVertexNormals();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 8)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("al.obj");
      loader->loadMtl("al.mtl", mesh);
      loader->loadObj("al.obj", mesh);
      mesh->normalize();
      mesh->computePerVertexNormals();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 9)
    {
      if (mesh) delete mesh;
      mesh = new FFmesh("teapot_five_spouts.obj");
      loader->loadMtl("teapot5s.mtl", mesh);
      loader->loadObj("teapot5s.obj", mesh);
      mesh->normalize();
      mesh->computePerVertexNormals();
      if (step_codec) delete step_codec;
      step_codec = 0;
      if (full_codec) delete full_codec;
      full_codec = encoder->encode(mesh);
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      aSlider->SetValue(0.0);
      aSlider->SetMax(full_codec->ls->size()+1);
    }
  else if (value == 10)
    {
      VertexNormalsOn = 0;
    }
  else if (value == 11)
    {
      VertexNormalsOn = 1;
    }
  else if (value == 12)
    {
      VertexNormalsOn = 2;
    }
  else if (value == 13)
    {
      VertexNormalsOn = 3;
    }
  else if (value == 14)
    {
      mesh->computePerVertexNormals();
    }
  else if (value == 15)
    {
      mesh->computePerFaceNormals();
    }
  else if (value == 20)
    {
      LightingAnimationOn =! LightingAnimationOn;
    }
  else if (value == 21)
    {
      LightCount = 1;
      InitLights();
    }
  else if (value == 22)
    {
      LightCount = 2;
      InitLights();
    }
  else if (value == 23)
    {
      LightCount = 3;
      InitLights();
    }
  else if (value == 24)
    {
      LightCount = 4;
      InitLights();
    }
  else if (value == 25)
    {
      LightCount = 5;
      InitLights();
    }
  else if (value == 26)
    {
      LightCount = 6;
      InitLights();
    }
  else if (value == 27)
    {
      LightCount = 7;
      InitLights();
    }
  else if (value == 30)
    {
      ColourMode = 0; // per material
    }
  else if (value == 31)
    {
      ColourMode = 1; // per group
    }
  else if (value == 32)
    {
      ColourMode = 2; // per component
    }
  else if (value == 33)
    {
      ColourMode = 3; // show triangles
    }
  else if (value == 34)
    {
      ColourMode = 4; // show quadrangles
    }
  else if (value == 35)
    {
      ColourMode = 5; // show pentagons
    }
  else if (value == 36)
    {
      ColourMode = 6; // show hexagons
    }
  else if (value == 37)
    {
      ColourMode = 7; // show biggergons
    }
  else if (value == 40)
    {
      WireFrameOn =! WireFrameOn;
    }
  else if (value == 41)
    {
      ViewMode = 0; // view entire model
    }
  else if (value == 42)
    {
      ViewMode = 1; // view single group
    }
  else if (value == 43)
    {
      ViewMode = 2; // view single component
    }
  else if (value == 44)
    {
      if (ViewMode == 1)
	{
	  ViewModeGroup = 1; // view first group
	}
      else if (ViewMode == 2)
	{
	  ViewModeComponent = 1; // view first component
	}
    }
  else if (value == 45)
    {
      if (ViewMode == 1)
	{
	  ViewModeGroup++; // view next group
	  if (ViewModeGroup >= mesh->gs->size())
	    {
	      ViewModeGroup = 1;
	    }
	}
      else if (ViewMode == 2)
	{
	  ViewModeComponent++; // view next component
	  if (ViewModeComponent > mesh->NumComponents)
	    {
	      ViewModeComponent = 1;
	    }
	}
    }
  else if (value == 51)
    {
      encodeMeshRel(1);
    }
  else if (value == 52)
    {
      encodeMeshRel(5);
    }
  else if (value == 53)
    {
      encodeMeshRel(10);
    }
  else if (value == 54)
    {
      CodingAnimationOn = 1;
      CodingAnimationSpeed = 0;
    }
  else if (value == 55)
    {
      CodingAnimationOn = 1;
      CodingAnimationSpeed = 1;
    }
  else if (value == 56)
    {
      CodingAnimationOn = 1;
      CodingAnimationSpeed = 2;
    }
  else if (value == 57)
    {
      CodingAnimationOn = 0;
    }
  else if (value == 58)
    {
      CodingAnimationOn = 0;
      CodingAnimationSpeed = 0;
      FFtwinEdge* e;
      if (step_codec) delete step_codec;
      step_codec = 0;
      for (int i = 0; i < mesh->es->size(); i++)
	{
	  e = (FFtwinEdge*) mesh->es->elementAt(i);
	  e->marker = 0;
	  e->boundary_prev = 0;
	  e->boundary_next = 0;
	}
      aSlider->SetValue(0.0);
    }
  else if (value == 100)
    {
      InteractionMode = 0; // rotate
    }
  else if (value == 101)
    {
      InteractionMode = 1; // translate
    }
  else if (value == 102)
    {
      InteractionMode = 2; // zoom      
    }
  else if (value == 103)
    {
      FILE* f_name = fopen("./labels.txt","w");
      if (f_name==NULL)
	{
	  cout << "ERROR: unable to open file: ./labels.txt" << endl;
	  f_name = stdout;
	}
      fprintf(f_name,"# %d Face Fixer labels to encode '%s' \n", full_codec->ls->size(), mesh->name);
      FFlabel* label;
      for (int i = 0; i < full_codec->ls->size(); i++)
	{
	  label = (FFlabel*) full_codec->ls->elementAt(i);
	  if (label->type == FFlabel::F)
	    {
	      fprintf(f_name,"F %d ", label->getDegree());
	    }
	  else if (label->type == FFlabel::H)
	    {
	      fprintf(f_name,"H %d ", label->getDegree());
	    }
	  else if (label->type == FFlabel::R)
	    {
	      fprintf(f_name,"R ");
	    }
	  else if (label->type == FFlabel::L)
	    {
	      fprintf(f_name,"L ");
	    }
	  else if (label->type == FFlabel::S)
	    {
	      fprintf(f_name,"S ");
	    }
	  else if (label->type == FFlabel::E)
	    {
	      fprintf(f_name,"E ");
	    }
	  else if (label->type == FFlabel::M)
	    {
	      fprintf(f_name,"M %d %d %d ", label->getIndex(), label->getOffset1(), label->getOffset2());
	    }
	}
      fprintf(f_name,"\n# end");
      fclose(f_name);
    }
  else if (value == 104)
    {
      Elevation = 0;
      Azimuth = 0;
      DistX = 0;
      DistY = 0;
      DistZ = 3;
      ViewMode = 0;
      ViewModeGroup = 1;
      ViewModeComponent = 1;
      InteractionMode = 0;
      LightingAnimationOn = 0;
      LightCount = 5;
      WireFrameOn = 0;
      VertexNormalsOn = 0;
      InteractionMode = 0;
      ColourMode = 0;
      MyMenuFunc(58);
      InitLights();
   }
  else if (value == 105)
    {
      exit(0);
    }
}

void myKeyboard(unsigned char Key, int x, int y)
{
  switch(Key)
    {
    case 27 :
    case 'Q':
    case 'q': exit(0);
    case ' ': // rotate, translate, or zoom
      if (InteractionMode == 2)
	{
	  InteractionMode = 0;
	}
      else
	{
	  InteractionMode += 1;
	}      
      break;
    case 'R':
    case 'r': // reset everything
      MyMenuFunc(104);
      break;
    case ',': //zoom in
      DistZ-=0.1f*DistZ;
      break;
    case '.': //zoom out
      DistZ+=0.1f*DistZ;
      break;
    case 'G':
    case 'g': // view single <g>roup
      if (ViewMode == 1)
	{
	  MyMenuFunc(41);
	}
      else
	{
	  MyMenuFunc(42);
	}
      break;
    case 'T':
    case 't': // encode step by step plus 1
      MyMenuFunc(51);
      break;
    case 'Y':
    case 'y': // encode step by step plus 5
      MyMenuFunc(52);
      break;
    case 'U':
    case 'u': // encode step by step plus 10
      MyMenuFunc(53);
      break;
    case 'P':
    case 'p': // play encode animation at different speeds
      if (CodingAnimationOn)
	{
	  if (CodingAnimationSpeed == 2)
	    {
	      CodingAnimationSpeed = 0;
	    }
	  else
	    {
	      CodingAnimationSpeed++;
	    }
	}
      else
	{
	  CodingAnimationOn = 1;
	}
      break;
    case 'O':
    case 'o': // stop encode animation
      CodingAnimationOn = 0;
      break;
    case 'I':
    case 'i': // reset encode animation
      MyMenuFunc(58);
      break;
    case 'C':
    case 'c': // view single <c>omponent
      if (ViewMode == 2)
	{
	  MyMenuFunc(41);
	}
      else
	{
	  MyMenuFunc(43);
	}
      break;
    case 'W':
    case 'w': // view <w>ireframe
      MyMenuFunc(40);
      break;
    case 'D':
    case 'd': // <d>ump
      MyMenuFunc(103);
      break;
    case 'F':
    case 'f': // view <f>irst
      MyMenuFunc(44);
      break;
    case 'V':
    case 'v':
      if (VertexNormalsOn == 3)
	{
	  VertexNormalsOn = 0;
	}
      else
	{
	  VertexNormalsOn += 1;
	}
      break;
    case 'N':
    case 'n': // view <n>ext
      MyMenuFunc(45);
      break;
    case 'S':
    case 's': // u<s>e materials
      MyMenuFunc(30);
      break;
    case 'E':
    case 'e': // random p<e>r group
      MyMenuFunc(31);
      break;
    case 'M':
    case 'm': // random per co<m>ponent
      MyMenuFunc(32);
      break;
    case 'Z':
    case 'z': // red triangles <z>
      MyMenuFunc(33);
      break;
    case 'X':
    case 'x': // red quadrangles <x>
      MyMenuFunc(34);
      break;
    case 'H':
    case 'h': // red pentagons <h>
      MyMenuFunc(35);
      break;
    case 'J':
    case 'j': // red hexagons <j>
      MyMenuFunc(36);
      break;
    case 'K':
    case 'k': // red biggergons <k>
      MyMenuFunc(37);
      break;
    case 'L':
    case 'l': // use 1-7 <l>ights
      MyMenuFunc(21 + ((LightCount)%7));
      break;
    case 'A':
    case 'a': // <a>nimate lights
      MyMenuFunc(20);
      break;
    case '1':
      MyMenuFunc(1);
      break;
    case '2':
      MyMenuFunc(2);
      break;
    case '3':
      MyMenuFunc(3);
      break;
    case '4':
      MyMenuFunc(4);
      break;
    case '5':
      MyMenuFunc(5);
      break;
    case '6':
      MyMenuFunc(6);
      break;
    case '7':
      MyMenuFunc(7);
      break;
    case '8':
      MyMenuFunc(8);
      break;
    case '9':
      MyMenuFunc(9);
      break;
  };
}

void InitLightModel()
{
  float GlobalAmbient[] = {0.0f, 0.3f, 0.0f, 1.0f};

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,GlobalAmbient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
}

void InitMaterials()
{
  colours_ambient[0][0] = 0.0f; colours_ambient[0][1] = 0.0f; colours_ambient[0][2] = 0.0f; colours_ambient[0][3] = 1.0f; // black
  colours_ambient[1][0] = 0.3f; colours_ambient[1][1] = 0.0f; colours_ambient[1][2] = 0.0f; colours_ambient[1][3] = 1.0f; // red
  colours_ambient[2][0] = 0.0f; colours_ambient[2][1] = 0.3f; colours_ambient[2][2] = 0.0f; colours_ambient[2][3] = 1.0f; // green
  colours_ambient[3][0] = 0.0f; colours_ambient[3][1] = 0.0f; colours_ambient[3][2] = 0.3f; colours_ambient[3][3] = 1.0f; // blue
  colours_ambient[4][0] = 0.3f; colours_ambient[4][1] = 0.3f; colours_ambient[4][2] = 0.0f; colours_ambient[4][3] = 1.0f; // yellow
  colours_ambient[5][0] = 0.3f; colours_ambient[5][1] = 0.0f; colours_ambient[5][2] = 0.3f; colours_ambient[5][3] = 1.0f; // purple
  colours_ambient[6][0] = 0.0f; colours_ambient[6][1] = 0.3f; colours_ambient[6][2] = 0.3f; colours_ambient[6][3] = 1.0f; // cyan
  colours_ambient[7][0] = 0.3f; colours_ambient[7][1] = 0.3f; colours_ambient[7][2] = 0.3f; colours_ambient[7][3] = 1.0f; // white
  colours_ambient[8][0] = 0.2f; colours_ambient[8][1] = 0.2f; colours_ambient[8][2] = 0.3f; colours_ambient[8][3] = 1.0f; // light blue
  colours_ambient[9][0] = 0.4f; colours_ambient[9][1] = 0.2f; colours_ambient[9][2] = 0.3f; colours_ambient[9][3] = 1.0f; // violett

  colours_diffuse[0][0] = 0.0f; colours_diffuse[0][1] = 0.0f; colours_diffuse[0][2] = 0.0f; colours_diffuse[0][3] = 1.0f; // black
  colours_diffuse[1][0] = 0.6f; colours_diffuse[1][1] = 0.0f; colours_diffuse[1][2] = 0.0f; colours_diffuse[1][3] = 1.0f; // red
  colours_diffuse[2][0] = 0.0f; colours_diffuse[2][1] = 0.6f; colours_diffuse[2][2] = 0.0f; colours_diffuse[2][3] = 1.0f; // green
  colours_diffuse[3][0] = 0.0f; colours_diffuse[3][1] = 0.0f; colours_diffuse[3][2] = 0.6f; colours_diffuse[3][3] = 1.0f; // blue
  colours_diffuse[4][0] = 0.6f; colours_diffuse[4][1] = 0.6f; colours_diffuse[4][2] = 0.0f; colours_diffuse[4][3] = 1.0f; // yellow
  colours_diffuse[5][0] = 0.6f; colours_diffuse[5][1] = 0.0f; colours_diffuse[5][2] = 0.6f; colours_diffuse[5][3] = 1.0f; // purple
  colours_diffuse[6][0] = 0.0f; colours_diffuse[6][1] = 0.6f; colours_diffuse[6][2] = 0.6f; colours_diffuse[6][3] = 1.0f; // cyan
  colours_diffuse[7][0] = 0.6f; colours_diffuse[7][1] = 0.6f; colours_diffuse[7][2] = 0.6f; colours_diffuse[7][3] = 1.0f; // white
  colours_diffuse[8][0] = 0.2f; colours_diffuse[8][1] = 0.2f; colours_diffuse[8][2] = 0.6f; colours_diffuse[8][3] = 1.0f; // light blue
  colours_diffuse[9][0] = 0.9f; colours_diffuse[9][1] = 0.4f; colours_diffuse[9][2] = 0.7f; colours_diffuse[9][3] = 1.0f; // violett

  colours_specular[0][0] = 0.0f; colours_specular[0][1] = 0.0f; colours_specular[0][2] = 0.0f; colours_specular[0][3] = 1.0f; // none
  colours_specular[1][0] = 0.3f; colours_specular[1][1] = 0.3f; colours_specular[1][2] = 0.3f; colours_specular[1][3] = 1.0f; // little
  colours_specular[2][0] = 0.7f; colours_specular[2][1] = 0.7f; colours_specular[2][2] = 0.7f; colours_specular[2][3] = 1.0f; // fairly
  colours_specular[3][0] = 1.0f; colours_specular[3][1] = 1.0f; colours_specular[3][2] = 1.0f; colours_specular[3][3] = 1.0f; // full

  colours_emission[0][0] = 0.0f; colours_emission[0][1] = 0.0f; colours_emission[0][2] = 0.0f; colours_emission[0][3] = 0.0f; // none
  colours_emission[1][0] = 0.5f; colours_emission[1][1] = 0.0f; colours_emission[1][2] = 0.0f; colours_emission[1][3] = 0.0f; // little red
  colours_emission[2][0] = 0.0f; colours_emission[2][1] = 0.5f; colours_emission[2][2] = 0.0f; colours_emission[2][3] = 0.0f; // little green
  colours_emission[3][0] = 0.0f; colours_emission[3][1] = 0.0f; colours_emission[3][2] = 0.5f; colours_emission[3][3] = 0.0f; // little blue
  colours_emission[4][0] = 0.4f; colours_emission[4][1] = 0.2f; colours_emission[4][2] = 0.3f; colours_emission[4][3] = 1.0f; // violett

  glMaterialf(GL_FRONT,GL_SHININESS,50);
}

void main(void)
{
  loader = new FFloader();
  encoder = new FFencoder();
  
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(WindowW,WindowH);
  glutInitWindowPosition(180,100);
  glutCreateWindow("SIGGRAPH 2000              Face Fixer              SIGGRAPH 2000");

  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  
  InitLightModel();
  InitLights();
  InitMaterials();

  glutDisplayFunc(myDisplay);
  glutReshapeFunc(myReshape);
  glutIdleFunc(myIdle);

  glutMouseFunc(myMouseFunc);
  glutMotionFunc(myMotionFunc);
  glutKeyboardFunc(myKeyboard);

  // models sub menu
  int menuModels = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("<1> teapot", 1);
  glutAddMenuEntry("<2> cow", 2);
  glutAddMenuEntry("<3> triceratops", 3);
  glutAddMenuEntry("<4> galleon", 4);
  glutAddMenuEntry("<5> shark", 5);
  glutAddMenuEntry("<6> beethoven", 6);
  glutAddMenuEntry("<7> cupie", 7);
  glutAddMenuEntry("<8> al", 8);
  glutAddMenuEntry("<9> five-spouted teapot", 9);

  // normals sub menu
  int menuNormals = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("show none <V>", 10);
  glutAddMenuEntry("show all <V>", 11);
  glutAddMenuEntry("show single <V>", 12);
  glutAddMenuEntry("show multiple <V>", 13);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("per vertex ", 14);
  glutAddMenuEntry("per face", 15);

  // lights sub menu
  int menuLights = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("<A>nimate", 20);
  glutAddMenuEntry("use 1 <L>ight", 21);
  glutAddMenuEntry("use 2 <L>ights", 22);
  glutAddMenuEntry("use 3 <L>ights", 23);
  glutAddMenuEntry("use 4 <L>ights", 24);
  glutAddMenuEntry("use 5 <L>ights", 25);
  glutAddMenuEntry("use 6 <L>ights", 26);
  glutAddMenuEntry("use 7 <L>ights", 27);

  // colours sub menu
  int menuColours = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("u<S>e materials", 30);
  glutAddMenuEntry("random p<E>r group", 31);
  glutAddMenuEntry("random per co<M>ponent", 32);
  glutAddMenuEntry("red triangles <Z>", 33);
  glutAddMenuEntry("red quadrangles <X>", 34);
  glutAddMenuEntry("red pentagons <H>", 35);
  glutAddMenuEntry("red hexagons <J>", 36);
  glutAddMenuEntry("red biggergons <K>", 37);

  // view sub menu
  int menuViews = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("view <W>ireframe", 40);
  glutAddMenuEntry("view entire model", 41);
  glutAddMenuEntry("view single <G>roup", 42);
  glutAddMenuEntry("view single <C>omponent", 43);
  glutAddMenuEntry("view <F>irst", 44);
  glutAddMenuEntry("view <N>ext", 45);

  // view sub menu
  int menuEncode = glutCreateMenu(MyMenuFunc);
  glutAddMenuEntry("step one <T>", 51);
  glutAddMenuEntry("step five <Y>", 52);
  glutAddMenuEntry("step ten <U>", 53);
  glutAddMenuEntry("play slow <P>", 54);
  glutAddMenuEntry("play medium <P>", 55);
  glutAddMenuEntry("play fast <P>", 56);
  glutAddMenuEntry("stop <O>", 57);
  glutAddMenuEntry("reset <I>", 58);

  // main menu
  glutCreateMenu(MyMenuFunc);
  glutAddSubMenu("models", menuModels);
  glutAddSubMenu("normals", menuNormals);
  glutAddSubMenu("lights", menuLights);
  glutAddSubMenu("colours", menuColours);
  glutAddSubMenu("views", menuViews);
  glutAddSubMenu("encode", menuEncode);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("rotate", 100);
  glutAddMenuEntry("translate", 101);
  glutAddMenuEntry("zoom", 102);
  glutAddMenuEntry("", 0);
  glutAddMenuEntry("<D>ump", 103);
  glutAddMenuEntry("<R>eset", 104);
  glutAddMenuEntry("<Q>UIT", 105);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  aSlider = new TUISlider(0, WindowH-30, WindowW, 30);

  MyMenuFunc(1);

  glutMainLoop();
}
