/*
===============================================================================

	FILE:	slider.h
	
	CONTENTS:
	
		Code for a simple slider in OpenGL
	
	PROGRAMMERS:
	
		Scott Cooper, derived from the GLOW library written by
		Daniel Azuma (DA)  <dazuma@kagi.com>
		Martin Isenburg, simplified from Scott Coopers hack
	
	COPYRIGHT:
	
		Copyright (C) 2000  Scott Cooper (scooper@cs.unc.edu)
		Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
		
		This library is free software; you can redistribute it and/or
		modify it under the terms of the GNU Lesser General Public
		License as published by the Free Software Foundation; either
		version 2.1 of the License, or (at your option) any later version.
		
		This library is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
		GNU Lesser General Public License for more details.
		
		You should have received a copy of the GNU Lesser General Public
		License along with this library; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
	
	CHANGE HISTORY:
	
		27 April 2000 -- SC Hacked mercilessly for Martin's
				    SIGGRAPH submission.
		28 April 2000 -- MI Hacked more merciless all night.
	
===============================================================================
*/
#ifndef TUI_SLIDER_H
#define TUI_SLIDER_H

class TUISlider
{  
public:
  
  TUISlider(const int x, const int y, const int width, const int height);
  TUISlider(const int x, const int y, const int width, const int height, const float min, const float max);
  
  void SetMin(const float min);
  float GetMin() { return fMin; }

  void SetMax(const float max);
  float GetMax() { return fMax; }
  
  int IsActive() { return 1; }
  int Height() { return fHeight; }
  int Width() { return fWidth; }

  int IsInside(int x, int y);
  int IsDragging();
  
  virtual void OnPaint();
  virtual void OnMouseDown(int button, int x, int y);
  virtual void OnMouseUp(int button, int x, int y);
  virtual void OnMouseMove(int x, int y);
  virtual void OnReleased();
  virtual void OnDragged();
  
  void DoDisplay();

  void SetValue(const float max); 
  float GetValue() { return fValue; }
  
protected:

  void CalcValue(int x, int y);

private:

  int fButtonPressed;
  int fDragging;
  int fNumTicks;
  float fValue;
  float fMin;
  float fMax;
  int fXOffset;
  int fYOffset;
  
  int fWidth;
  int fHeight;
  int fXPosition;
  int fYPosition;
  
  int fTop;
  int fBottom;
  int fLeft;
  int fRight;

  GLint fOldScissor[4];
};

#endif
